//	Copyright (C) 1999, 2000 Core Technologies.
//
//	This file is part of tpasm.
//
//	tpasm is free software; you can redistribute it and/or modify
//	it under the terms of the tpasm LICENSE AGREEMENT.
//
//	tpasm is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	tpasm LICENSE AGREEMENT for more details.
//
//	You should have received a copy of the tpasm LICENSE AGREEMENT
//	along with tpasm; see the file "LICENSE.TXT".


// Generate code for Motorola 6805

#include	"include.h"

static SYMTABLE
	*opcodeSymbols;

// enumerated addressing modes

#define	OT_INHERENT				0								// no operands
#define	OT_IMMEDIATE			1								// one byte immediate operand
#define	OT_DIRECT				2								// one byte direct
#define	OT_EXTENDED				3								// two byte absolute
#define	OT_INDIRECT_2			4								// indirect with 2 bytes of offset
#define	OT_INDIRECT_1			5								// indirect with 1 byte of offset
#define	OT_INDIRECT_0			6								// indirect with 0 bytes of offset
#define	OT_RELATIVE				7								// one byte relative offset
#define	OT_BIT_DIRECT			8								// (convenience) one byte of direct, one relative offset
#define	OT_DIRECT_RELATIVE		9								// one byte of direct, one relative offset
#define	OT_BIT_DIRECT_RELATIVE	10								// (convenience) one bit index, one byte of direct, one relative offset

#define	OT_NUM					OT_BIT_DIRECT_RELATIVE+1		// number of addressing modes

// masks for the various addressing modes

#define	M_INHERENT				(1<<OT_INHERENT)
#define	M_IMMEDIATE				(1<<OT_IMMEDIATE)
#define	M_DIRECT				(1<<OT_DIRECT)
#define	M_EXTENDED				(1<<OT_EXTENDED)
#define	M_INDIRECT_2			(1<<OT_INDIRECT_2)
#define	M_INDIRECT_1			(1<<OT_INDIRECT_1)
#define	M_INDIRECT_0			(1<<OT_INDIRECT_0)
#define	M_RELATIVE				(1<<OT_RELATIVE)
#define	M_BIT_DIRECT			(1<<OT_BIT_DIRECT)
#define	M_DIRECT_RELATIVE		(1<<OT_DIRECT_RELATIVE)
#define	M_BIT_DIRECT_RELATIVE	(1<<OT_BIT_DIRECT_RELATIVE)

typedef bool HandlerFunction(struct opcode *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord);

typedef struct opcode
{
	char
		*name;
	HandlerFunction
		*pseudoHandler;
	int
		typeMask;
	unsigned char
		baseOpcode[OT_NUM];
} OPCODE;

// prototypes for the handler functions which need to be declared before we can build
// the opcode table

static bool HandleDB(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord);
static bool HandleDW(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord);
static bool HandleDS(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord);
static bool HandleDSW(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord);
static bool HandleIncbin(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord);

#define	____		0					// unused opcode

static OPCODE
	Opcodes[]=
	{
		// pseudo ops

		{"db",		HandleDB,		0,																			{____,____,____,____,____,____,____,____,____,____,____}},
		{"dc.b",	HandleDB,		0,																			{____,____,____,____,____,____,____,____,____,____,____}},
		{"dw",		HandleDW,		0,																			{____,____,____,____,____,____,____,____,____,____,____}},
		{"dc.w",	HandleDW,		0,																			{____,____,____,____,____,____,____,____,____,____,____}},
		{"ds",		HandleDS,		0,																			{____,____,____,____,____,____,____,____,____,____,____}},
		{"ds.b",	HandleDS,		0,																			{____,____,____,____,____,____,____,____,____,____,____}},
		{"ds.w",	HandleDSW,		0,																			{____,____,____,____,____,____,____,____,____,____,____}},
		{"incbin",	HandleIncbin,	0,																			{____,____,____,____,____,____,____,____,____,____,____}},

		// opcodes
//																												inh   imm  dir  ext  ind2 ind1 ind0 rel  bd   dr   bdr
		{"adc",		NULL,		M_IMMEDIATE|M_DIRECT|M_EXTENDED|M_INDIRECT_2|M_INDIRECT_1|M_INDIRECT_0,			{____,0xA9,0xB9,0xC9,0xD9,0xE9,0xF9,____,____,____,____}},
		{"add",		NULL,		M_IMMEDIATE|M_DIRECT|M_EXTENDED|M_INDIRECT_2|M_INDIRECT_1|M_INDIRECT_0,			{____,0xAB,0xBB,0xCB,0xDB,0xEB,0xFB,____,____,____,____}},
		{"and",		NULL,		M_IMMEDIATE|M_DIRECT|M_EXTENDED|M_INDIRECT_2|M_INDIRECT_1|M_INDIRECT_0,			{____,0xA4,0xB4,0xC4,0xD4,0xE4,0xF4,____,____,____,____}},
		{"asl",		NULL,		M_DIRECT|M_INDIRECT_1|M_INDIRECT_0,												{____,____,0x38,____,____,0x68,0x78,____,____,____,____}},
		{"asla",	NULL,		M_INHERENT,																		{0x48,____,____,____,____,____,____,____,____,____,____}},
		{"aslx",	NULL,		M_INHERENT,																		{0x58,____,____,____,____,____,____,____,____,____,____}},
		{"asr",		NULL,		M_DIRECT|M_INDIRECT_1|M_INDIRECT_0,												{____,____,0x37,____,____,0x67,0x77,____,____,____,____}},
		{"asra",	NULL,		M_INHERENT,																		{0x47,____,____,____,____,____,____,____,____,____,____}},
		{"asrx",	NULL,		M_INHERENT,																		{0x57,____,____,____,____,____,____,____,____,____,____}},
		{"bcc",		NULL,		M_RELATIVE,																		{____,____,____,____,____,____,____,0x24,____,____,____}},
		{"bclr",	NULL,		M_BIT_DIRECT,																	{____,____,____,____,____,____,____,____,0x11,____,____}},
		{"bclr0",	NULL,		M_DIRECT,																		{____,____,0x11,____,____,____,____,____,____,____,____}},
		{"bclr1",	NULL,		M_DIRECT,																		{____,____,0x13,____,____,____,____,____,____,____,____}},
		{"bclr2",	NULL,		M_DIRECT,																		{____,____,0x15,____,____,____,____,____,____,____,____}},
		{"bclr3",	NULL,		M_DIRECT,																		{____,____,0x17,____,____,____,____,____,____,____,____}},
		{"bclr4",	NULL,		M_DIRECT,																		{____,____,0x19,____,____,____,____,____,____,____,____}},
		{"bclr5",	NULL,		M_DIRECT,																		{____,____,0x1B,____,____,____,____,____,____,____,____}},
		{"bclr6",	NULL,		M_DIRECT,																		{____,____,0x1D,____,____,____,____,____,____,____,____}},
		{"bclr7",	NULL,		M_DIRECT,																		{____,____,0x1F,____,____,____,____,____,____,____,____}},
		{"bcs",		NULL,		M_RELATIVE,																		{____,____,____,____,____,____,____,0x25,____,____,____}},
		{"beq",		NULL,		M_RELATIVE,																		{____,____,____,____,____,____,____,0x27,____,____,____}},
		{"bhcc",	NULL,		M_RELATIVE,																		{____,____,____,____,____,____,____,0x28,____,____,____}},
		{"bhcs",	NULL,		M_RELATIVE,																		{____,____,____,____,____,____,____,0x29,____,____,____}},
		{"bhi",		NULL,		M_RELATIVE,																		{____,____,____,____,____,____,____,0x22,____,____,____}},
		{"bhs",		NULL,		M_RELATIVE,																		{____,____,____,____,____,____,____,0x24,____,____,____}},
		{"bih",		NULL,		M_RELATIVE,																		{____,____,____,____,____,____,____,0x2F,____,____,____}},
		{"bil",		NULL,		M_RELATIVE,																		{____,____,____,____,____,____,____,0x2E,____,____,____}},
		{"bit",		NULL,		M_IMMEDIATE|M_DIRECT|M_EXTENDED|M_INDIRECT_2|M_INDIRECT_1|M_INDIRECT_0,			{____,0xA5,0xB5,0xC5,0xD5,0xE5,0xF5,____,____,____,____}},
		{"blo",		NULL,		M_RELATIVE,																		{____,____,____,____,____,____,____,0x25,____,____,____}},
		{"bls",		NULL,		M_RELATIVE,																		{____,____,____,____,____,____,____,0x23,____,____,____}},
		{"bmc",		NULL,		M_RELATIVE,																		{____,____,____,____,____,____,____,0x2C,____,____,____}},
		{"bmi",		NULL,		M_RELATIVE,																		{____,____,____,____,____,____,____,0x2B,____,____,____}},
		{"bms",		NULL,		M_RELATIVE,																		{____,____,____,____,____,____,____,0x2D,____,____,____}},
		{"bne",		NULL,		M_RELATIVE,																		{____,____,____,____,____,____,____,0x26,____,____,____}},
		{"bpl",		NULL,		M_RELATIVE,																		{____,____,____,____,____,____,____,0x2A,____,____,____}},
		{"bra",		NULL,		M_RELATIVE,																		{____,____,____,____,____,____,____,0x20,____,____,____}},
		{"brclr",	NULL,		M_BIT_DIRECT_RELATIVE,															{____,____,____,____,____,____,____,____,____,____,0x01}},
		{"brclr0",	NULL,		M_DIRECT_RELATIVE,																{____,____,____,____,____,____,____,____,____,0x01,____}},
		{"brclr1",	NULL,		M_DIRECT_RELATIVE,																{____,____,____,____,____,____,____,____,____,0x03,____}},
		{"brclr2",	NULL,		M_DIRECT_RELATIVE,																{____,____,____,____,____,____,____,____,____,0x05,____}},
		{"brclr3",	NULL,		M_DIRECT_RELATIVE,																{____,____,____,____,____,____,____,____,____,0x07,____}},
		{"brclr4",	NULL,		M_DIRECT_RELATIVE,																{____,____,____,____,____,____,____,____,____,0x09,____}},
		{"brclr5",	NULL,		M_DIRECT_RELATIVE,																{____,____,____,____,____,____,____,____,____,0x0B,____}},
		{"brclr6",	NULL,		M_DIRECT_RELATIVE,																{____,____,____,____,____,____,____,____,____,0x0D,____}},
		{"brclr7",	NULL,		M_DIRECT_RELATIVE,																{____,____,____,____,____,____,____,____,____,0x0F,____}},
		{"brset",	NULL,		M_BIT_DIRECT_RELATIVE,															{____,____,____,____,____,____,____,____,____,____,0x00}},
		{"brset0",	NULL,		M_DIRECT_RELATIVE,																{____,____,____,____,____,____,____,____,____,0x00,____}},
		{"brset1",	NULL,		M_DIRECT_RELATIVE,																{____,____,____,____,____,____,____,____,____,0x02,____}},
		{"brset2",	NULL,		M_DIRECT_RELATIVE,																{____,____,____,____,____,____,____,____,____,0x04,____}},
		{"brset3",	NULL,		M_DIRECT_RELATIVE,																{____,____,____,____,____,____,____,____,____,0x06,____}},
		{"brset4",	NULL,		M_DIRECT_RELATIVE,																{____,____,____,____,____,____,____,____,____,0x08,____}},
		{"brset5",	NULL,		M_DIRECT_RELATIVE,																{____,____,____,____,____,____,____,____,____,0x0A,____}},
		{"brset6",	NULL,		M_DIRECT_RELATIVE,																{____,____,____,____,____,____,____,____,____,0x0C,____}},
		{"brset7",	NULL,		M_DIRECT_RELATIVE,																{____,____,____,____,____,____,____,____,____,0x0E,____}},
		{"brn",		NULL,		M_RELATIVE,																		{____,____,____,____,____,____,____,0x21,____,____,____}},
		{"bset",	NULL,		M_BIT_DIRECT,																	{____,____,____,____,____,____,____,____,0x10,____,____}},
		{"bset0",	NULL,		M_DIRECT,																		{____,____,0x10,____,____,____,____,____,____,____,____}},
		{"bset1",	NULL,		M_DIRECT,																		{____,____,0x12,____,____,____,____,____,____,____,____}},
		{"bset2",	NULL,		M_DIRECT,																		{____,____,0x14,____,____,____,____,____,____,____,____}},
		{"bset3",	NULL,		M_DIRECT,																		{____,____,0x16,____,____,____,____,____,____,____,____}},
		{"bset4",	NULL,		M_DIRECT,																		{____,____,0x18,____,____,____,____,____,____,____,____}},
		{"bset5",	NULL,		M_DIRECT,																		{____,____,0x1A,____,____,____,____,____,____,____,____}},
		{"bset6",	NULL,		M_DIRECT,																		{____,____,0x1C,____,____,____,____,____,____,____,____}},
		{"bset7",	NULL,		M_DIRECT,																		{____,____,0x1E,____,____,____,____,____,____,____,____}},
		{"bsr",		NULL,		M_RELATIVE,																		{____,____,____,____,____,____,____,0xAD,____,____,____}},
		{"clc",		NULL,		M_INHERENT,																		{0x98,____,____,____,____,____,____,____,____,____,____}},
		{"cli",		NULL,		M_INHERENT,																		{0x9A,____,____,____,____,____,____,____,____,____,____}},
		{"clr",		NULL,		M_DIRECT|M_INDIRECT_1|M_INDIRECT_0,												{____,____,0x3F,____,____,0x6F,0x7F,____,____,____,____}},
		{"clra",	NULL,		M_INHERENT,																		{0x4F,____,____,____,____,____,____,____,____,____,____}},
		{"clrx",	NULL,		M_INHERENT,																		{0x5F,____,____,____,____,____,____,____,____,____,____}},
		{"cmp",		NULL,		M_IMMEDIATE|M_DIRECT|M_EXTENDED|M_INDIRECT_2|M_INDIRECT_1|M_INDIRECT_0,			{____,0xA1,0xB1,0xC1,0xD1,0xE1,0xF1,____,____,____,____}},
		{"com",		NULL,		M_DIRECT|M_INDIRECT_1|M_INDIRECT_0,												{____,____,0x33,____,____,0x63,0x73,____,____,____,____}},
		{"coma",	NULL,		M_INHERENT,																		{0x43,____,____,____,____,____,____,____,____,____,____}},
		{"comx",	NULL,		M_INHERENT,																		{0x53,____,____,____,____,____,____,____,____,____,____}},
		{"cpx",		NULL,		M_IMMEDIATE|M_DIRECT|M_EXTENDED|M_INDIRECT_2|M_INDIRECT_1|M_INDIRECT_0,			{____,0xA3,0xB3,0xC3,0xD3,0xE3,0xF3,____,____,____,____}},
		{"dec",		NULL,		M_DIRECT|M_INDIRECT_1|M_INDIRECT_0,												{____,____,0x3A,____,____,0x6A,0x7A,____,____,____,____}},
		{"deca",	NULL,		M_INHERENT,																		{0x4A,____,____,____,____,____,____,____,____,____,____}},
		{"decx",	NULL,		M_INHERENT,																		{0x5A,____,____,____,____,____,____,____,____,____,____}},
		{"dex",		NULL,		M_INHERENT,																		{0x5A,____,____,____,____,____,____,____,____,____,____}},
		{"eor",		NULL,		M_IMMEDIATE|M_DIRECT|M_EXTENDED|M_INDIRECT_2|M_INDIRECT_1|M_INDIRECT_0,			{____,0xA8,0xB8,0xC8,0xD8,0xE8,0xF8,____,____,____,____}},
		{"inc",		NULL,		M_DIRECT|M_INDIRECT_1|M_INDIRECT_0,												{____,____,0x3C,____,____,0x6C,0x7C,____,____,____,____}},
		{"inca",	NULL,		M_INHERENT,																		{0x4C,____,____,____,____,____,____,____,____,____,____}},
		{"incx",	NULL,		M_INHERENT,																		{0x5C,____,____,____,____,____,____,____,____,____,____}},
		{"inx",		NULL,		M_INHERENT,																		{0x5C,____,____,____,____,____,____,____,____,____,____}},
		{"jmp",		NULL,		M_DIRECT|M_EXTENDED|M_INDIRECT_2|M_INDIRECT_1|M_INDIRECT_0,						{____,____,0xBC,0xCC,0xDC,0xEC,0xFC,____,____,____,____}},
		{"jsr",		NULL,		M_DIRECT|M_EXTENDED|M_INDIRECT_2|M_INDIRECT_1|M_INDIRECT_0,						{____,____,0xBD,0xCD,0xDD,0xED,0xFD,____,____,____,____}},
		{"lda",		NULL,		M_IMMEDIATE|M_DIRECT|M_EXTENDED|M_INDIRECT_2|M_INDIRECT_1|M_INDIRECT_0,			{____,0xA6,0xB6,0xC6,0xD6,0xE6,0xF6,____,____,____,____}},
		{"ldx",		NULL,		M_IMMEDIATE|M_DIRECT|M_EXTENDED|M_INDIRECT_2|M_INDIRECT_1|M_INDIRECT_0,			{____,0xAE,0xBE,0xCE,0xDE,0xEE,0xFE,____,____,____,____}},
		{"lsl",		NULL,		M_DIRECT|M_INDIRECT_1|M_INDIRECT_0,												{____,____,0x38,____,____,0x68,0x78,____,____,____,____}},
		{"lsla",	NULL,		M_INHERENT,																		{0x48,____,____,____,____,____,____,____,____,____,____}},
		{"lslx",	NULL,		M_INHERENT,																		{0x58,____,____,____,____,____,____,____,____,____,____}},
		{"lsr",		NULL,		M_DIRECT|M_INDIRECT_1|M_INDIRECT_0,												{____,____,0x34,____,____,0x64,0x74,____,____,____,____}},
		{"lsra",	NULL,		M_INHERENT,																		{0x44,____,____,____,____,____,____,____,____,____,____}},
		{"lsrx",	NULL,		M_INHERENT,																		{0x54,____,____,____,____,____,____,____,____,____,____}},
		{"mul",		NULL,		M_INHERENT,																		{0x42,____,____,____,____,____,____,____,____,____,____}},
		{"neg",		NULL,		M_DIRECT|M_INDIRECT_1|M_INDIRECT_0,												{____,____,0x30,____,____,0x60,0x70,____,____,____,____}},
		{"nega",	NULL,		M_INHERENT,																		{0x40,____,____,____,____,____,____,____,____,____,____}},
		{"negx",	NULL,		M_INHERENT,																		{0x50,____,____,____,____,____,____,____,____,____,____}},
		{"nop",		NULL,		M_INHERENT,																		{0x9D,____,____,____,____,____,____,____,____,____,____}},
		{"ora",		NULL,		M_IMMEDIATE|M_DIRECT|M_EXTENDED|M_INDIRECT_2|M_INDIRECT_1|M_INDIRECT_0,			{____,0xAA,0xBA,0xCA,0xDA,0xEA,0xFA,____,____,____,____}},
		{"rol",		NULL,		M_DIRECT|M_INDIRECT_1|M_INDIRECT_0,												{____,____,0x39,____,____,0x69,0x79,____,____,____,____}},
		{"rola",	NULL,		M_INHERENT,																		{0x49,____,____,____,____,____,____,____,____,____,____}},
		{"rolx",	NULL,		M_INHERENT,																		{0x59,____,____,____,____,____,____,____,____,____,____}},
		{"ror",		NULL,		M_DIRECT|M_INDIRECT_1|M_INDIRECT_0,												{____,____,0x36,____,____,0x66,0x76,____,____,____,____}},
		{"rora",	NULL,		M_INHERENT,																		{0x46,____,____,____,____,____,____,____,____,____,____}},
		{"rorx",	NULL,		M_INHERENT,																		{0x56,____,____,____,____,____,____,____,____,____,____}},
		{"rsp",		NULL,		M_INHERENT,																		{0x9C,____,____,____,____,____,____,____,____,____,____}},
		{"rti",		NULL,		M_INHERENT,																		{0x80,____,____,____,____,____,____,____,____,____,____}},
		{"rts",		NULL,		M_INHERENT,																		{0x81,____,____,____,____,____,____,____,____,____,____}},
		{"sbc",		NULL,		M_IMMEDIATE|M_DIRECT|M_EXTENDED|M_INDIRECT_2|M_INDIRECT_1|M_INDIRECT_0,			{____,0xA2,0xB2,0xC2,0xD2,0xE2,0xF2,____,____,____,____}},
		{"sec",		NULL,		M_INHERENT,																		{0x99,____,____,____,____,____,____,____,____,____,____}},
		{"sei",		NULL,		M_INHERENT,																		{0x9B,____,____,____,____,____,____,____,____,____,____}},
		{"sta",		NULL,		M_DIRECT|M_EXTENDED|M_INDIRECT_2|M_INDIRECT_1|M_INDIRECT_0,						{____,____,0xB7,0xC7,0xD7,0xE7,0xF7,____,____,____,____}},
		{"stop",	NULL,		M_INHERENT,																		{0x8E,____,____,____,____,____,____,____,____,____,____}},
		{"stx",		NULL,		M_DIRECT|M_EXTENDED|M_INDIRECT_2|M_INDIRECT_1|M_INDIRECT_0,						{____,____,0xBF,0xCF,0xDF,0xEF,0xFF,____,____,____,____}},
		{"sub",		NULL,		M_IMMEDIATE|M_DIRECT|M_EXTENDED|M_INDIRECT_2|M_INDIRECT_1|M_INDIRECT_0,			{____,0xA0,0xB0,0xC0,0xD0,0xE0,0xF0,____,____,____,____}},
		{"swi",		NULL,		M_INHERENT,																		{0x83,____,____,____,____,____,____,____,____,____,____}},
		{"tax",		NULL,		M_INHERENT,																		{0x97,____,____,____,____,____,____,____,____,____,____}},
		{"tst",		NULL,		M_DIRECT|M_INDIRECT_1|M_INDIRECT_0,												{____,____,0x3D,____,____,0x6D,0x7D,____,____,____,____}},
		{"tsta",	NULL,		M_INHERENT,																		{0x4D,____,____,____,____,____,____,____,____,____,____}},
		{"tstx",	NULL,		M_INHERENT,																		{0x5D,____,____,____,____,____,____,____,____,____,____}},
		{"txa",		NULL,		M_INHERENT,																		{0x9F,____,____,____,____,____,____,____,____,____,____}},
		{"wait",	NULL,		M_INHERENT,																		{0x8F,____,____,____,____,____,____,____,____,____,____}},
	};

static bool GenerateByte(unsigned char value,LISTING_RECORD *listingRecord)
// output the value to the current segment
// This will return false only if a "hard" error occurs
{
	bool
		fail;
	int
		length;

	fail=false;
	if(currentSegment)
	{
		if(currentSegment->currentPC<65536)
		{
			if(!intermediatePass)				// only do the real work if necessary
			{
				length=strlen(listingRecord->listObjectString);
				if(length+2<MAXSTRING)
				{
					sprintf(&listingRecord->listObjectString[length],"%02X ",value);		// create list file output
				}
				fail=!AddBytesToSegment(currentSegment,currentSegment->currentPC,&value,1);
			}
			currentSegment->currentPC++;
		}
		else
		{
			AssemblyComplaint(NULL,true,"Attempt to write code outside address space\n");
		}
	}
	else
	{
		AssemblyComplaint(NULL,true,"Code cannot occur outside of a segment\n");
	}
	return(!fail);
}

static bool ParseOperandSeparator(char *theLine,int *lineIndex)
// Expect a comma, step over one if found
{
	if(theLine[*lineIndex]==',')			// does this look like an opcode separator?
	{
		(*lineIndex)++;						// step over it
		return(true);
	}
	return(false);
}

static bool ParseImmediatePreamble(char *theLine,int *lineIndex)
// Expect a pound sign, step over one if found
{
	if(theLine[*lineIndex]=='#')			// does this look like a '#'?
	{
		(*lineIndex)++;						// step over it
		return(true);
	}
	return(false);
}

static bool ParseIndex(char *theLine,int *lineIndex)
// see if next thing on the line is an index register
// return true if so, false otherwise
{
	if(theLine[*lineIndex]=='X'||theLine[*lineIndex]=='x')
	{
		if(!IsLabelChar(theLine[(*lineIndex)+1]))
		{
			(*lineIndex)++;						// step over the register
			return(true);
		}
	}
	return(false);
}

// opcode and processor-specific pseudo opcode handling for 6805

// pseudo op handling

static bool HandleDB(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// declaring bytes of data
{
	int
		value;
	bool
		unresolved;
	char
		outputString[MAXSTRING];
	int
		stringLength;
	bool
		done,
		fail;
	int
		i;

	fail=false;
	done=false;
	while(!done&&!fail)
	{
		if(ParseQuotedString(theLine,lineIndex,'"','"',outputString,&stringLength))
		{
			i=0;
			while(!fail&&i<stringLength)
			{
				fail=!GenerateByte(outputString[i],listingRecord);
				i++;
			}
		}
		else if(ParseExpression(theLine,lineIndex,&value,&unresolved))
		{
			CheckByteRange(value,true,true);
			fail=!GenerateByte(value&0xFF,listingRecord);
		}
		else
		{
			ReportBadOperands();
			done=true;
		}
		if(!done&&!fail)				// look for separator, or comment
		{
			if(ParseComment(theLine,lineIndex))
			{
				done=true;
			}
			else if(ParseOperandSeparator(theLine,lineIndex))
			{
			}
			else
			{
				ReportBadOperands();
				done=true;
			}
		}
	}
	return(!fail);
}

static bool HandleDW(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// declaring words of data in big endian
{
	int
		value;
	bool
		unresolved;
	char
		outputString[MAXSTRING];
	int
		stringLength;
	bool
		done,
		fail;
	int
		i;

	fail=false;
	done=false;
	while(!done&&!fail)
	{
		if(ParseQuotedString(theLine,lineIndex,'"','"',outputString,&stringLength))
		{
			i=0;
			while(!fail&&i<stringLength)
			{
				if(GenerateByte(0,listingRecord))
				{
					fail=!GenerateByte(outputString[i],listingRecord);
				}
				else
				{
					fail=true;
				}
				i++;
			}
		}
		else if(ParseExpression(theLine,lineIndex,&value,&unresolved))
		{
			CheckWordRange(value,true,true);
			if(GenerateByte(value>>8,listingRecord))
			{
				fail=!GenerateByte(value&0xFF,listingRecord);
			}
			else
			{
				fail=true;
			}
		}
		else
		{
			ReportBadOperands();
			done=true;
		}
		if(!done&&!fail)				// look for separator, or comment
		{
			if(ParseComment(theLine,lineIndex))
			{
				done=true;
			}
			else if(ParseOperandSeparator(theLine,lineIndex))
			{
			}
			else
			{
				ReportBadOperands();
				done=true;
			}
		}
	}
	return(!fail);
}

static bool HandleDS(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// move the PC forward by the given amount
{
	int
		value;
	bool
		unresolved;

	if(ParseExpression(theLine,lineIndex,&value,&unresolved))
	{
		if(ParseComment(theLine,lineIndex))
		{
			if(currentSegment)
			{
				if(!unresolved)
				{
					currentSegment->currentPC+=value;
				}
			}
			else
			{
				AssemblyComplaint(NULL,true,"'%s' outside of segment\n",opcode->name);
			}
		}
		else
		{
			ReportBadOperands();
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(true);
}

static bool HandleDSW(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// move the PC forward by the given amount times two (do as if skipping this many words)
{
	int
		value;
	bool
		unresolved;

	if(ParseExpression(theLine,lineIndex,&value,&unresolved))
	{
		if(ParseComment(theLine,lineIndex))
		{
			value*=2;
			if(currentSegment)
			{
				if(!unresolved)
				{
					currentSegment->currentPC+=value;
				}
			}
			else
			{
				AssemblyComplaint(NULL,true,"'%s' outside of segment\n",opcode->name);
			}
		}
		else
		{
			ReportBadOperands();
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(true);
}

static bool HandleIncbin(OPCODE *opcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// include the data of the given file into the input stream
// If there is a problem, report it and return false
{
	bool
		fail;
	char
		fileName[MAXSTRING];
	int
		stringLength;
	FILE
		*theFile;
	char
		inputDataBuffer[1024];
	int
		bytesRead;
	int
		i;

	fail=false;

	if(ParseQuotedString(theLine,lineIndex,'"','"',fileName,&stringLength))
	{
		if(ParseComment(theLine,lineIndex))				// make sure there's nothing else on the line
		{
			if((theFile=fopen(fileName,"rb")))
			{
				while(!fail&&((bytesRead=fread(inputDataBuffer,1,1024,theFile))>0))
				{
					for(i=0;i<bytesRead;i++)
					{
						fail=!GenerateByte(inputDataBuffer[i],listingRecord);	// ### this is really inefficient
					}
				}
				fclose(theFile);
			}
			else
			{
				AssemblyComplaint(NULL,true,"Failed to open file '%s'. %s\n",fileName,strerror(errno));
			}
		}
		else
		{
			AssemblyComplaint(NULL,true,"Missing or incorrectly formatted file name\n");
		}
	}
	else
	{
		AssemblyComplaint(NULL,true,"Missing or incorrectly formatted file name\n");
	}
	return(!fail);
}

// opcode handling

enum
{
	POT_INDEX,
	POT_IMMEDIATE,
	POT_VALUE,
};

static bool ParseOperandElement(char *theLine,int *lineIndex,int *type,int *value,bool *unresolved)
// Try to parse an element of the operand and determine its type
// return true if the parsing succeeds
{
	if(ParseImmediatePreamble(theLine,lineIndex))		// an immediate operand?
	{
		if(ParseExpression(theLine,lineIndex,value,unresolved))
		{
			*type=POT_IMMEDIATE;
			return(true);
		}
	}
	else if(ParseIndex(theLine,lineIndex))
	{
		*type=POT_INDEX;
		return(true);
	}
	else if(ParseExpression(theLine,lineIndex,value,unresolved))
	{
		*type=POT_VALUE;
		return(true);
	}
	return(false);
}

static bool HandleImmediate(OPCODE *theOpcode,int value,bool unresolved,LISTING_RECORD *listingRecord)
// an immediate addressing mode was located
{
	bool
		fail;

	fail=false;
	if(theOpcode->typeMask&M_IMMEDIATE)
	{
		CheckByteRange(value,true,true);
		if(GenerateByte(theOpcode->baseOpcode[OT_IMMEDIATE],listingRecord))
		{
			fail=!GenerateByte(value,listingRecord);
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleDirect(OPCODE *theOpcode,int value,bool unresolved,LISTING_RECORD *listingRecord)
// deal with direct mode output only
{
	bool
		fail;

	fail=false;
	if(theOpcode->typeMask&M_DIRECT)
	{
		CheckUnsignedByteRange(value,true,true);
		if(GenerateByte(theOpcode->baseOpcode[OT_DIRECT],listingRecord))
		{
			fail=!GenerateByte(value,listingRecord);
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleExtended(OPCODE *theOpcode,int value,bool unresolved,LISTING_RECORD *listingRecord)
// deal with extended mode output only
{
	bool
		fail;

	fail=false;
	if(theOpcode->typeMask&M_EXTENDED)
	{
		CheckUnsignedWordRange(value,true,true);
		if(GenerateByte(theOpcode->baseOpcode[OT_EXTENDED],listingRecord))
		{
			if(GenerateByte(value>>8,listingRecord))
			{
				fail=!GenerateByte(value&0xFF,listingRecord);
			}
			else
			{
				fail=true;
			}
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleDirectOrExtended(OPCODE *theOpcode,int value,bool unresolved,LISTING_RECORD *listingRecord)
// a direct or extended address has been parsed
{
	bool
		fail;

	fail=false;
	if(theOpcode->typeMask&M_DIRECT)
	{
		if(((value>=0)&&(value<256))||!(theOpcode->typeMask&M_EXTENDED))
		{
			fail=!HandleDirect(theOpcode,value,unresolved,listingRecord);
		}
		else
		{
			fail=!HandleExtended(theOpcode,value,unresolved,listingRecord);
		}
	}
	else
	{
		fail=!HandleExtended(theOpcode,value,unresolved,listingRecord);
	}
	return(!fail);
}

static bool HandleIndirect0(OPCODE *theOpcode,LISTING_RECORD *listingRecord)
// an indirect 0 addressing mode was located
{
	bool
		fail;

	fail=false;
	if(theOpcode->typeMask&M_INDIRECT_0)
	{
		fail=!GenerateByte(theOpcode->baseOpcode[OT_INDIRECT_0],listingRecord);
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleIndirect1(OPCODE *theOpcode,int value,bool unresolved,LISTING_RECORD *listingRecord)
// an indirect 1 addressing mode was located
{
	bool
		fail;

	fail=false;
	if(theOpcode->typeMask&M_INDIRECT_1)
	{
		CheckUnsignedByteRange(value,true,true);
		if(GenerateByte(theOpcode->baseOpcode[OT_INDIRECT_1],listingRecord))
		{
			fail=!GenerateByte(value,listingRecord);
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleIndirect2(OPCODE *theOpcode,int value,bool unresolved,LISTING_RECORD *listingRecord)
// an indirect 2 addressing mode was located
{
	bool
		fail;

	fail=false;
	if(theOpcode->typeMask&M_INDIRECT_2)
	{
		CheckUnsignedWordRange(value,true,true);
		if(GenerateByte(theOpcode->baseOpcode[OT_INDIRECT_2],listingRecord))
		{
			if(GenerateByte(value>>8,listingRecord))
			{
				fail=!GenerateByte(value&0xFF,listingRecord);
			}
			else
			{
				fail=true;
			}
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleIndirect1or2(OPCODE *theOpcode,int value,bool unresolved,LISTING_RECORD *listingRecord)
// an indirect 1 or 2 addressing mode was located
{
	bool
		fail;

	fail=false;
	if(theOpcode->typeMask&M_INDIRECT_1)
	{
		if(((value>=0)&&(value<256))||!(theOpcode->typeMask&M_INDIRECT_2))
		{
			fail=!HandleIndirect1(theOpcode,value,unresolved,listingRecord);
		}
		else
		{
			fail=!HandleIndirect2(theOpcode,value,unresolved,listingRecord);
		}
	}
	else
	{
		fail=!HandleIndirect2(theOpcode,value,unresolved,listingRecord);
	}
	return(!fail);
}

static bool HandleBitDirect(OPCODE *theOpcode,int value1,bool unresolved1,int value2,bool unresolved2,LISTING_RECORD *listingRecord)
// a bit direct operand has been parsed
{
	bool
		fail;

	fail=false;
	if(theOpcode->typeMask&M_BIT_DIRECT)
	{
		Check8BitIndexRange(value1,true,true);
		CheckUnsignedByteRange(value2,true,true);
		value1&=0x07;
		if(GenerateByte(theOpcode->baseOpcode[OT_BIT_DIRECT]|(value1<<1),listingRecord))
		{
			fail=!GenerateByte(value2,listingRecord);
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleDirectRelative(OPCODE *theOpcode,int value1,bool unresolved1,int value2,bool unresolved2,LISTING_RECORD *listingRecord)
// a direct relative operand has been parsed
{
	int
		offset;
	bool
		fail;

	fail=false;
	if(theOpcode->typeMask&M_DIRECT_RELATIVE)
	{
		if(GenerateByte(theOpcode->baseOpcode[OT_DIRECT_RELATIVE],listingRecord))
		{
			CheckUnsignedByteRange(value1,true,true);
			if(GenerateByte(value1,listingRecord))
			{
				if(unresolved2)
				{
					offset=0;
				}
				else
				{
					if(currentSegment)
					{
						offset=value2-(currentSegment->currentPC+currentSegment->codeGenOffset)-1;
					}
					else
					{
						offset=0;
					}
					Check8RelativeRange(offset,true,true);
				}
				fail=!GenerateByte(offset,listingRecord);
			}
			else
			{
				fail=true;
			}
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleBitDirectRelative(OPCODE *theOpcode,int value1,bool unresolved1,int value2,bool unresolved2,int value3,bool unresolved3,LISTING_RECORD *listingRecord)
// a bit direct relative operand has been parsed
{
	int
		offset;
	bool
		fail;

	fail=false;
	if(theOpcode->typeMask&M_BIT_DIRECT_RELATIVE)
	{
		Check8BitIndexRange(value1,true,true);
		value1&=0x07;
		if(GenerateByte(theOpcode->baseOpcode[OT_BIT_DIRECT_RELATIVE]|(value1<<1),listingRecord))
		{
			CheckUnsignedByteRange(value2,true,true);
			if(GenerateByte(value2,listingRecord))
			{
				if(unresolved3)
				{
					offset=0;
				}
				else
				{
					if(currentSegment)
					{
						offset=value3-(currentSegment->currentPC+currentSegment->codeGenOffset)-1;
					}
					else
					{
						offset=0;
					}
					Check8RelativeRange(offset,true,true);
				}
				fail=!GenerateByte(offset,listingRecord);
			}
			else
			{
				fail=true;
			}
		}
		else
		{
			fail=true;
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleRelative(OPCODE *theOpcode,int value,bool unresolved,LISTING_RECORD *listingRecord)
// a relative operand has been parsed
{
	bool
		fail;
	int
		offset;

	fail=false;
	if(GenerateByte(theOpcode->baseOpcode[OT_RELATIVE],listingRecord))
	{
		if(unresolved)
		{
			offset=0;
		}
		else
		{
			if(currentSegment)
			{
				offset=value-(currentSegment->currentPC+currentSegment->codeGenOffset)-1;
			}
			else
			{
				offset=0;
			}
			Check8RelativeRange(offset,true,true);
		}
		fail=!GenerateByte(offset,listingRecord);
	}
	else
	{
		fail=true;
	}
	return(!fail);
}

static bool HandleSingleAddress(OPCODE *theOpcode,int value,bool unresolved,LISTING_RECORD *listingRecord)
// a lone address has been parsed
// This means it is either a relative address, direct, or extended
{
	bool
		fail;

	fail=false;
	if(theOpcode->typeMask&M_RELATIVE)
	{
		fail=!HandleRelative(theOpcode,value,unresolved,listingRecord);
	}
	else if(theOpcode->typeMask&(M_DIRECT|M_EXTENDED))
	{
		fail=!HandleDirectOrExtended(theOpcode,value,unresolved,listingRecord);
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleDoubleAddress(OPCODE *theOpcode,int value1,bool unresolved1,int value2,bool unresolved2,LISTING_RECORD *listingRecord)
// a dual address has been parsed
// This means it is either BIT_DIRECT, or DIRECT_RELATIVE
{
	bool
		fail;

	fail=false;
	if(theOpcode->typeMask&M_BIT_DIRECT)
	{
		fail=!HandleBitDirect(theOpcode,value1,unresolved1,value2,unresolved2,listingRecord);
	}
	else if(theOpcode->typeMask&M_DIRECT_RELATIVE)
	{
		fail=!HandleDirectRelative(theOpcode,value1,unresolved1,value2,unresolved2,listingRecord);
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleTripleAddress(OPCODE *theOpcode,int value1,bool unresolved1,int value2,bool unresolved2,int value3,bool unresolved3,LISTING_RECORD *listingRecord)
// a triple address has been parsed
// This means it is BIT_DIRECT_RELATIVE
{
	bool
		fail;

	fail=false;
	if(theOpcode->typeMask&M_BIT_DIRECT_RELATIVE)
	{
		fail=!HandleBitDirectRelative(theOpcode,value1,unresolved1,value2,unresolved2,value3,unresolved3,listingRecord);
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleSecondValue(OPCODE *theOpcode,char *theLine,int *lineIndex,int value1,bool unresolved1,int value2,bool unresolved2,LISTING_RECORD *listingRecord)
// there are two value, and something still remains
// the only 3 valued operand type is BIT_DIRECT_RELATIVE
// so parse and apply
{
	bool
		fail;
	int
		elementType;
	int
		value3;
	bool
		unresolved3;

	fail=false;
	if(ParseOperandSeparator(theLine,lineIndex))		// make sure the next thing separates operands
	{
		if(ParseOperandElement(theLine,lineIndex,&elementType,&value3,&unresolved3))
		{
			if(elementType==POT_VALUE)
			{
				if(ParseComment(theLine,lineIndex))
				{
					return(HandleTripleAddress(theOpcode,value1,unresolved1,value2,unresolved2,value3,unresolved3,listingRecord));
				}
				else
				{
					ReportBadOperands();
				}
			}
			else
			{
				ReportBadOperands();
			}
		}
		else
		{
			ReportBadOperands();
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleFirstValue(OPCODE *theOpcode,char *theLine,int *lineIndex,int value1,bool unresolved1,LISTING_RECORD *listingRecord)
// a non-immediate value was parsed, so see what else we can find
{
	bool
		fail;
	int
		elementType;
	int
		value2;
	bool
		unresolved2;

	fail=false;
	if(ParseOperandSeparator(theLine,lineIndex))		// make sure the next thing separates operands
	{
		if(ParseOperandElement(theLine,lineIndex,&elementType,&value2,&unresolved2))
		{
			if(elementType==POT_INDEX)
			{
				if(ParseComment(theLine,lineIndex))
				{
					return(HandleIndirect1or2(theOpcode,value1,unresolved1,listingRecord));
				}
				else
				{
					ReportBadOperands();
				}
			}
			else if(elementType==POT_VALUE)
			{
				if(ParseComment(theLine,lineIndex))
				{
					return(HandleDoubleAddress(theOpcode,value1,unresolved1,value2,unresolved2,listingRecord));
				}
				else
				{
					return(HandleSecondValue(theOpcode,theLine,lineIndex,value1,unresolved1,value2,unresolved2,listingRecord));
				}
			}
			else
			{
				ReportBadOperands();
			}
		}
		else
		{
			ReportBadOperands();
		}
	}
	else
	{
		ReportBadOperands();
	}
	return(!fail);
}

static bool HandleOpcode(void *theOpcode,char *theLine,int *lineIndex,LISTING_RECORD *listingRecord)
// look at the type of opcode available, parse operands as allowed
// return false only if there was a 'hard' error
{
	OPCODE
		*actualOpcode;
	bool
		fail;
	int
		elementType;
	int
		value;
	bool
		unresolved;

	fail=false;
	actualOpcode=(OPCODE *)theOpcode;
	if(actualOpcode->pseudoHandler)
	{
		fail=!actualOpcode->pseudoHandler(actualOpcode,theLine,lineIndex,listingRecord);
	}
	else
	{
		if(!ParseComment(theLine,lineIndex))
		{
			if(ParseOperandSeparator(theLine,lineIndex))	// see if initial separator (allowed if next thing is an index register)
			{
				if(ParseOperandElement(theLine,lineIndex,&elementType,&value,&unresolved))
				{
					if(elementType==POT_INDEX)
					{
						if(ParseComment(theLine,lineIndex))
						{
							return(HandleIndirect0(actualOpcode,listingRecord));
						}
						else
						{
							ReportBadOperands();
						}
					}
					else
					{
						ReportBadOperands();
					}
				}
				else
				{
					ReportBadOperands();
				}
			}
			else	// first character of operand was not a ','
			{
				if(ParseOperandElement(theLine,lineIndex,&elementType,&value,&unresolved))
				{
					if(elementType==POT_INDEX)
					{
						if(ParseComment(theLine,lineIndex))
						{
							return(HandleIndirect0(actualOpcode,listingRecord));
						}
						else
						{
							ReportBadOperands();
						}
					}
					else if(elementType==POT_IMMEDIATE)
					{
						if(ParseComment(theLine,lineIndex))
						{
							return(HandleImmediate(actualOpcode,value,unresolved,listingRecord));
						}
						else
						{
							ReportBadOperands();
						}
					}
					else if(elementType==POT_VALUE)
					{
						if(ParseComment(theLine,lineIndex))
						{
							return(HandleSingleAddress(actualOpcode,value,unresolved,listingRecord));
						}
						else
						{
							return(HandleFirstValue(actualOpcode,theLine,lineIndex,value,unresolved,listingRecord));
						}
					}
				}
				else
				{
					ReportBadOperands();
				}
			}
		}
		else
		{
			if(actualOpcode->typeMask&M_INHERENT)
			{
				fail=!GenerateByte(actualOpcode->baseOpcode[OT_INHERENT],listingRecord);
			}
			else
			{
				ReportBadOperands();
			}
		}
	}
	return(!fail);
}

static void *MatchOpcode(char *theOpcode)
// match opcodes for this processor, return NULL if none matched
{
	return(STFindDataForName(opcodeSymbols,theOpcode));
}

static bool SelectProcessor(PROCESSOR *theProcessor)
// A processor in this family is being selected to assemble with
{
	return(true);
}

static void DeselectProcessor(PROCESSOR *theProcessor)
// A processor in this family is being deselected
{
}

static void UnInitFamily()
// undo what InitFamily did
{
	STDisposeSymbolTable(opcodeSymbols);
}

static bool InitFamily()
// initialize symbol table
{
	unsigned int
		i;
	bool
		fail;

	fail=false;
	if((opcodeSymbols=STNewSymbolTable()))
	{
		for(i=0;!fail&&(i<sizeof(Opcodes)/sizeof(OPCODE));i++)
		{
			if(!STAddEntryAtEnd(opcodeSymbols,Opcodes[i].name,&Opcodes[i]))
			{
				fail=true;
			}
		}
		if(!fail)
		{
			return(true);
		}
		STDisposeSymbolTable(opcodeSymbols);
	}
	return(false);
}

// processors handled here (the constuctors for these variables link them to the global
// list of processors that the assembler knows how to handle)

static PROCESSOR_FAMILY
	processorFamily("Motorola 6805",InitFamily,UnInitFamily,SelectProcessor,DeselectProcessor,MatchOpcode,HandleOpcode);

static PROCESSOR
	processor0(&processorFamily,"6805",NULL),
	processor1(&processorFamily,"68705",NULL);
