//	Copyright (C) 1999, 2000, 2001 Core Technologies.
//
//	This file is part of tpasm.
//
//	tpasm is free software; you can redistribute it and/or modify
//	it under the terms of the tpasm LICENSE AGREEMENT.
//
//	tpasm is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	tpasm LICENSE AGREEMENT for more details.
//
//	You should have received a copy of the tpasm LICENSE AGREEMENT
//	along with tpasm; see the file "LICENSE.TXT".


// Support routines used by numerous processors to help keep the
// amount of duplicated code to a minimum

#include	"include.h"

void ReportBadOperands()
// complain that the operands are bad
{
	AssemblyComplaint(NULL,true,"Invalid operands\n");
}

bool CheckSignedByteRange(int theValue,bool generateMessage,bool isError)
// see if theValue looks like a signed byte
// if not, return false, and generate a warning or error if told to
{
	if(theValue>=-128&&theValue<128)
	{
		return(true);
	}
	if(generateMessage)
	{
		AssemblyComplaint(NULL,isError,"Signed byte value (0x%X) out of range\n",theValue);
	}
	return(false);
}

bool CheckUnsignedByteRange(int theValue,bool generateMessage,bool isError)
// see if theValue looks like an unsigned byte
// if not, return false, and generate a warning or error if told to
{
	if(theValue>=0&&theValue<256)
	{
		return(true);
	}
	if(generateMessage)
	{
		AssemblyComplaint(NULL,isError,"Unsigned byte value (0x%X) out of range\n",theValue);
	}
	return(false);
}

bool CheckByteRange(int theValue,bool generateMessage,bool isError)
// see if theValue looks like a signed byte, or an unsigned byte
// if not, return false, and generate a warning or error if told to
{
	if(theValue>=-128&&theValue<256)
	{
		return(true);
	}
	if(generateMessage)
	{
		AssemblyComplaint(NULL,isError,"Byte value (0x%X) out of range\n",theValue);
	}
	return(false);
}

bool CheckSignedWordRange(int theValue,bool generateMessage,bool isError)
// see if theValue looks like a signed word
// if not, return false, and generate a warning or error if told to
{
	if(theValue>=-32768&&theValue<32768)
	{
		return(true);
	}
	if(generateMessage)
	{
		AssemblyComplaint(NULL,isError,"Signed word value (0x%X) out of range\n",theValue);
	}
	return(false);
}

bool CheckUnsignedWordRange(int theValue,bool generateMessage,bool isError)
// see if theValue looks like an unsigned word
// if not, return false, and generate a warning or error if told to
{
	if(theValue>=0&&theValue<65536)
	{
		return(true);
	}
	if(generateMessage)
	{
		AssemblyComplaint(NULL,isError,"Unsigned word value (0x%X) out of range\n",theValue);
	}
	return(false);
}

bool CheckWordRange(int theValue,bool generateMessage,bool isError)
// see if theValue looks like a signed word, or an unsigned word
// if not, return false, and generate a warning or error if told to
{
	if(theValue>=-32768&&theValue<65536)
	{
		return(true);
	}
	if(generateMessage)
	{
		AssemblyComplaint(NULL,isError,"Word value (0x%X) out of range\n",theValue);
	}
	return(false);
}

bool Check8BitIndexRange(int theValue,bool generateMessage,bool isError)
// check value to see if it looks good as 8 bit index
// complain if not
{
	if(theValue>=0&&theValue<8)
	{
		return(true);
	}
	if(generateMessage)
	{
		AssemblyComplaint(NULL,isError,"Bit index (0x%X) out of range\n",theValue);
	}
	return(false);
}

bool Check16BitIndexRange(int theValue,bool generateMessage,bool isError)
// check value to see if it looks good as 16 bit index
// complain if not
{
	if(theValue>=0&&theValue<16)
	{
		return(true);
	}
	if(generateMessage)
	{
		AssemblyComplaint(NULL,isError,"Bit index (0x%X) out of range\n",theValue);
	}
	return(false);
}

bool Check32BitIndexRange(int theValue,bool generateMessage,bool isError)
// check value to see if it looks good as 32 bit index
// complain if not
{
	if(theValue>=0&&theValue<32)
	{
		return(true);
	}
	if(generateMessage)
	{
		AssemblyComplaint(NULL,isError,"Bit index (0x%X) out of range\n",theValue);
	}
	return(false);
}

bool Check8RelativeRange(int theValue,bool generateMessage,bool isError)
// check value to see if it looks good as 8 bit relative offset
// complain if not
{
	if(theValue>=-128&&theValue<128)
	{
		return(true);
	}
	if(generateMessage)
	{
		AssemblyComplaint(NULL,isError,"Relative offset (0x%X) out of range\n",theValue);
	}
	return(false);
}

bool Check16RelativeRange(int theValue,bool generateMessage,bool isError)
// check value to see if it looks good as 16 bit relative offset
// complain if not
{
	if(theValue>=-32768&&theValue<32768)
	{
		return(true);
	}
	if(generateMessage)
	{
		AssemblyComplaint(NULL,isError,"Relative offset (0x%X) out of range\n",theValue);
	}
	return(false);
}
